/*
Copyright (c) 2016 David Gow <david@ingeniumdigital.com> 

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE. 
*/
#include <SDL.h>
#include <GL/glew.h>
#include <GL/glu.h>

#include "map.h"
#include "sprite.h"

void LoadTex(const char *fname, GLuint val)
{
	SDL_Surface *surf = SDL_LoadBMP(fname);
	glBindTexture(GL_TEXTURE_2D, val);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, surf->w, surf->h, 0, GL_BGR, GL_UNSIGNED_BYTE, surf->pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	SDL_FreeSurface(surf);
}


void SetupCamera(float x, float y, float angle)
{
	glEnable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//glFrustum(-1, 1, -1, 1, 0.1, 5.0);
	gluPerspective(45.0,16.0/9.00,0.1,100.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(x,0,y,cos(angle) + x, 0, sin(angle) + y, 0, 1, 0);
	
}


void Draw()
{
	SetupCamera(player.x,player.y,player.angle);
	
	MAP_Draw();
	SPRITE_DrawAll();
	
}

enum GameState {
	STATE_OPENING,
	STATE_INSTRUCTION,
	STATE_INTERSTITIAL,
	STATE_GAMEOVER,
	STATE_WIN,
	STATE_PLAYING
};

GameState state;

GLuint screenTex[STATE_PLAYING];

void STATE_LoadAssets()
{
	glGenTextures(STATE_PLAYING, screenTex);
	LoadTex("open.bmp", screenTex[STATE_OPENING]);
	LoadTex("inst.bmp", screenTex[STATE_INSTRUCTION]);
	LoadTex("over.bmp", screenTex[STATE_GAMEOVER]);
	LoadTex("inter.bmp", screenTex[STATE_INTERSTITIAL]);
	LoadTex("end.bmp", screenTex[STATE_WIN]);
}

bool ScreenState()
{
	if (state == STATE_PLAYING)
		return true;
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glDisable(GL_CULL_FACE);
	glBindTexture(GL_TEXTURE_2D, screenTex[state]);
	glEnable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	glTexCoord2f(0.0,1.0);
	glVertex2f(-1.f,-1.f);
	glTexCoord2f(1.0,1.0);
	glVertex2f(1.0,-1.0);
	glTexCoord2f(1.0,0.0);
	glVertex2f(1.f,1.f);
	glTexCoord2f(0.0,0.0);
	glVertex2f(-1.f,1.f);
	glEnd();
	glDisable(GL_TEXTURE_2D);
}

bool gameover = false;

void nextState()
{
	switch (state)
	{
	case STATE_OPENING:
		state = STATE_INSTRUCTION;
		break;
	case STATE_WIN:
		gameover = true;
		break;
	default:
		SPRITE_InitMap();
		for (int y = 0; y < 22; ++y)
		{
			for (int x = 0; x < 16; ++x)
			{
				if (MAP_GetTile(x,y) == 'o')
				{
					SPRITE_Place(x, y, true);
				}
				else if (MAP_GetTile(x,y) == '^')
				{
					SPRITE_Place(x, y, false);
				}
				else if (MAP_GetTile(x,y) == 'P')
				{
					player = {x,y,0.15,3.14159265/2.0,1.0,1.0,1.0};
				}
			}
		}
		state = STATE_PLAYING;
	}
}


struct Input
{
	bool fwd;
	bool back;
	bool left;
	bool right;
	bool change_me;
	bool old_change_me;
};

int main(int argc, char **argv)
{
	SDL_Init(SDL_INIT_VIDEO);
	SDL_Window *wnd = SDL_CreateWindow("Shapeshift [sulix]", 0,0,1366,768,SDL_WINDOW_OPENGL);
	SDL_GLContext ctx = SDL_GL_CreateContext(wnd);
	glewInit();
	SDL_GL_SetSwapInterval(1);
	Input input = {0};
	MAP_LoadAssets();
	STATE_LoadAssets();
	SPRITE_LoadAssets();
	MAP_Set(0);
	
	while (!gameover)
	{
		SDL_Event evt;
		input.old_change_me = input.change_me;
		while (SDL_PollEvent(&evt))
		{
			switch (evt.type)
			{
			case SDL_QUIT:
				return 0;
			case SDL_KEYUP:
			case SDL_KEYDOWN:
				switch(evt.key.keysym.sym)
				{
				case SDLK_LEFT:
				case SDLK_a:
					input.left = (evt.type == SDL_KEYDOWN);
					break;
				case SDLK_RIGHT:
				case SDLK_d:
					input.right = (evt.type == SDL_KEYDOWN);
					break;
				case SDLK_UP:
				case SDLK_w:
					input.fwd = (evt.type == SDL_KEYDOWN);
					break;
				case SDLK_DOWN:
				case SDLK_s:
					input.back = (evt.type == SDL_KEYDOWN);
					break;
				case SDLK_SPACE:
					input.change_me = (evt.type == SDL_KEYDOWN);
					break;
				}
			}
		}
		glClearColor(1,1,1,1);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
		
		if (ScreenState())
		{
			if (input.fwd)
			{
				SPRITE_Thrust(&player, 0.1, player.angle);
			}
			if (input.back)
			{
				SPRITE_Thrust(&player, -0.1, player.angle);
			}
			if (input.left)
			{
				player.angle -= 0.0314;
			}
			if (input.right)
			{
				player.angle += 0.0314;
			}
			if (input.old_change_me && !input.change_me)
			{
				SPRITE_ToggleType(&player);
			}
		
			PLAYER_Update();
			SPRITE_DoAllCollisions();
			Draw();
			PLAYER_DrawOverlay();
			
			if (player.health <= 0.0) state = STATE_GAMEOVER;
			if (MAP_GetTile(round(player.x), round(player.y)) == '_')
			{
				//TODO: multiple maps
				state = STATE_INTERSTITIAL;
				if (MAP_Next())
					state = STATE_WIN;
			}
		}
		else
		{
			if (input.old_change_me && !input.change_me)
				nextState();
		}
		SDL_GL_SwapWindow(wnd);
	}
	SDL_Quit();
	return 0;
}
