/*
Copyright (c) 2016 David Gow <david@ingeniumdigital.com> 

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE. 
*/
#include <SDL.h>
#include <GL/glew.h>
#include <GL/glu.h>
#include <vector>
#include <stdio.h>

GLuint wallTex=0;
int map_width = 16, map_height = 22;

std::vector<char *> maps;

const char *map;
void MAP_LoadAssets()
{
	SDL_Surface *wall = SDL_LoadBMP("wall.bmp");
	glGenTextures(1, &wallTex);
	glBindTexture(GL_TEXTURE_2D, wallTex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, wall->w, wall->h, 0, GL_BGR, GL_UNSIGNED_BYTE, wall->pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	SDL_FreeSurface(wall);
	
	FILE *mapfile = fopen("maps.dat", "r");
	int numMaps;
	fscanf(mapfile,"%d\n", &numMaps);
	for (int m = 0; m < numMaps; ++m)
	{
		char *cmap = new char[16*22+1];
		for (int y = 0; y < 22; ++y)
		{
			for (int x = 0; x < 16; ++x)
			{
				fscanf(mapfile,"%c", &cmap[y*16+x]);
			}
			fscanf(mapfile,"\n");
		}
		maps.push_back(cmap);
	}
	fclose(mapfile);
}


void MAP_Set(int id)
{
	map = maps[id];
}

int mapnum = 0;

bool MAP_Next()
{
	mapnum++;
	if (mapnum < maps.size())
	{
		MAP_Set(mapnum);
		return false;
	}
	return true;
}


void MAP_DrawTile(int x, int y)
{
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, wallTex);
	float fx = float(x), fy = float(y);
	glBegin(GL_QUADS);
	glColor4f(0.5,0.5,0.5,1);
	glTexCoord2f(0, 0);
	glVertex3f(fx-0.5, -0.5, fy-0.5);
	glTexCoord2f(0, 1);
	glVertex3f(fx-0.5, 0.5, fy-0.5);
	glTexCoord2f(1, 1);
	glVertex3f(fx+0.5, 0.5, fy-0.5);
	glTexCoord2f(1, 0);
	glVertex3f(fx+0.5, -0.5, fy-0.5);
	
	glColor4f(0.8,0.8,0.8,1);
	glTexCoord2f(0, 0);
	glVertex3f(fx+0.5, -0.5, fy-0.5);
	glTexCoord2f(0, 1);
	glVertex3f(fx+0.5, 0.5, fy-0.5);
	glTexCoord2f(1, 1);
	glVertex3f(fx+0.5, 0.5, fy+0.5);
	glTexCoord2f(1, 0);
	glVertex3f(fx+0.5, -0.5, fy+0.5);
	
	glColor4f(0.5,0.5,0.5,1);
	glTexCoord2f(0, 0);
	glVertex3f(fx+0.5, -0.5, fy+0.5);
	glTexCoord2f(0, 1);
	glVertex3f(fx+0.5, 0.5, fy+0.5);
	glTexCoord2f(1, 1);
	glVertex3f(fx-0.5, 0.5, fy+0.5);
	glTexCoord2f(1, 0);
	glVertex3f(fx-0.5, -0.5, fy+0.5);
	
	glColor4f(0.8,0.8,0.8,1);
	glTexCoord2f(0, 0);
	glVertex3f(fx-0.5, -0.5, fy+0.5);
	glTexCoord2f(0, 1);
	glVertex3f(fx-0.5, 0.5, fy+0.5);
	glTexCoord2f(1, 1);
	glVertex3f(fx-0.5, 0.5, fy-0.5);
	glTexCoord2f(1, 0);
	glVertex3f(fx-0.5, -0.5, fy-0.5);
	glEnd();
	glDisable(GL_TEXTURE_2D);
}


// Returns the tile at pos (x,y)
char MAP_GetTile(int x, int y)
{
	SDL_assert(x < 16);
	SDL_assert(y < 22);
	return map[y*16+x];
}

bool MAP_IsSolid(int x, int y)
{
	return MAP_GetTile(x,y) == 'X';
}

bool MAP_LineIntersectsSolid(float x1, float y1, float x2, float y2)
{
	// Hacky raymarch.
	for (float lambda = 0.0; lambda < 1.0; lambda += 0.05)
	{
		float x_lerp = x1 + (x2 - x1)*lambda;
		float y_lerp = y1 + (y2 - y1)*lambda;
		int x_tile = round(x_lerp);
		int y_tile = round(y_lerp);
		if (MAP_IsSolid(x_tile, y_tile))
			return true;
	}
	return false;
}

void MAP_Draw()
{
	glColor4f(0,1,0,1);
	glBegin(GL_QUADS);
	glVertex3f(-100.0, -1.0, -100.0);
	glVertex3f(-100.0, -1.0, 100.0);
	glVertex3f(100.0, -1.0, 100.0);
	glVertex3f(100.0, -1.0, -100.0);
	glEnd();
	glColor4f(0,0,1,1);
	glBegin(GL_QUADS);
	glVertex3f(-100.0, 1.0, -100.0);
	glVertex3f(100.0, 1.0, -100.0);
	glVertex3f(100.0, 1.0, 100.0);
	glVertex3f(-100.0, 1.0, 100.0);
	glEnd();
	
	for (int y = 0; y < 22; ++y)
	{
		for (int x = 0; x < 16; ++x)
		{
			if (MAP_GetTile(x,y) == 'X')
			{
				MAP_DrawTile(x,y);
			}
			else if (MAP_GetTile(x,y) == '_')
			{
				glColor4f(1,1,0,1);
				glBegin(GL_QUADS);
				glVertex3f(x - 0.5, -0.5, y - 0.5);
				glVertex3f(x - 0.5, -0.5, y + 0.5);
				glVertex3f(x + 0.5, -0.5, y + 0.5);
				glVertex3f(x + 0.5, -0.5, y - 0.5);
				glEnd();
			}
		}
	}
}